package nn_assignment;

public class dataUnit
{
	public double[] values;
	public double label = 0.0;
	public int numberOfValues;

	//Constructor to create a blank dataUnit
	public dataUnit(int numberOfInputs)
	{
		numberOfValues = numberOfInputs;
		values = new double[numberOfInputs];
	}

	//Constructor to create a populated dataUnit
	public dataUnit(double[] inputArray, double inputLabel)
	{
		numberOfValues = inputArray.length;
		values = new double[inputArray.length];

		//Now read in all the values they've given us
		for (int i = 0; i < inputArray.length; i++)
		{
			values[i] = inputArray[i];
		}

		//Now set the type for this data
		label = inputLabel;
	}

}